@@ -1,12 +1,12 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps NFA_this_repr;
 
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st = MVM_gc_allocate_stable(tc, &NFA_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
@@ -79,16 +79,23 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
     /* Write number of states. */
     MVM_serialization_write_int(tc, writer, body->num_states);
 
-    /* Write state edge list counts. */
-    for (i = 0; i < body->num_states; i++)
-        MVM_serialization_write_int(tc, writer, body->num_state_edges[i]);
+    /* Write state edge list counts, skipping synthetic start node. */
+    for (i = 0; i < body->num_states; i++) {
+        MVMint64 sig_edges = body->num_state_edges[i];
+        if (sig_edges && body->states[i][0].act == MVM_NFA_EDGE_SYNTH_CP_COUNT)
+            sig_edges--;
+        MVM_serialization_write_int(tc, writer, sig_edges);
+    }
 
     /* Write state graph. */
     for (i = 0; i < body->num_states; i++) {
         for (j = 0; j < body->num_state_edges[i]; j++) {
-            MVM_serialization_write_int(tc, writer, body->states[i][j].act);
+            MVMint64 act = body->states[i][j].act;
+            if (act == MVM_NFA_EDGE_SYNTH_CP_COUNT)
+                continue;
+            MVM_serialization_write_int(tc, writer, act);
             MVM_serialization_write_int(tc, writer, body->states[i][j].to);
-            switch (body->states[i][j].act & 0xff) {
+            switch (act & 0xff) {
                 case MVM_NFA_EDGE_FATE:
                     MVM_serialization_write_int(tc, writer, body->states[i][j].arg.i);
                     break;
@@ -107,10 +114,9 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
                          * and then each of the codepoints. */
                         MVMNFGSynthetic *si = MVM_nfg_get_synthetic_info(tc, g);
                         MVMint32 k;
-                        MVM_serialization_write_int(tc, writer, -(si->num_combs + 1));
-                        MVM_serialization_write_int(tc, writer, si->base);
-                        for (k = 0; k < si->num_combs; k++)
-                            MVM_serialization_write_int(tc, writer, si->combs[k]);
+                        MVM_serialization_write_int(tc, writer, -(si->num_codes));
+                        for (k = 0; k < si->num_codes; k++)
+                            MVM_serialization_write_int(tc, writer, si->codes[k]);
                     }
                     break;
                 }
@@ -140,6 +146,70 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
     }
 }
 
+/* Go through each state and see if there are at least four edges involving
+ * a simple codepoint match. If so, sort them to the start and stick in a
+ * synthetic edge which indicates the number of codepoint edges ahead. This
+ * means that our matching can binary search with the codepoint it has, and
+ * skip over any inappropriate edges. */
+static int classify_edge(MVMNFAStateInfo *e) {
+    switch (e->act) {
+        case MVM_NFA_EDGE_SYNTH_CP_COUNT:
+            return 0;
+        case MVM_NFA_EDGE_CODEPOINT:
+        case MVM_NFA_EDGE_CODEPOINT_LL:
+            return 1;
+        default:
+            return 2;
+    }
+}
+static int opt_edge_comp(const void *av, const void *bv) {
+    MVMNFAStateInfo *a = (MVMNFAStateInfo *)av;
+    MVMNFAStateInfo *b = (MVMNFAStateInfo *)bv;
+    MVMint32 type_a = classify_edge(a);
+    MVMint32 type_b = classify_edge(b);
+    if (type_a < type_b)
+        return -1;
+    if (type_a > type_b)
+        return 1;
+    if (type_a == 1) {
+        return a->arg.g < b->arg.g ? -1 :
+               a->arg.g > b->arg.g ?  1 :
+                                      0;
+    }
+    else {
+        return 0;
+    }
+}
+static void sort_states_and_add_synth_cp_node(MVMThreadContext *tc, MVMNFABody *body) {
+    MVMint64 s;
+    for (s = 0; s < body->num_states; s++) {
+        /* See if there's enough interesting edges to do the opt. */
+        MVMint32 applicable_edges = 0;
+        MVMint64 num_orig_edges = body->num_state_edges[s];
+        if (num_orig_edges >= 4) {
+            MVMint64 e;
+            for (e = 0; e < num_orig_edges; e++) {
+                MVMint64 act = body->states[s][e].act;
+                if (act == MVM_NFA_EDGE_CODEPOINT || act == MVM_NFA_EDGE_CODEPOINT_LL)
+                    applicable_edges++;
+            }
+        }
+
+        /* If enough edges, insert synthetic and so the sort. */
+        if (applicable_edges >= 4) {
+            MVMint64 num_new_edges = num_orig_edges + 1;
+            MVMNFAStateInfo *new_edges = MVM_malloc(num_new_edges * sizeof(MVMNFAStateInfo));
+            new_edges[0].act = MVM_NFA_EDGE_SYNTH_CP_COUNT;
+            new_edges[0].arg.i = applicable_edges;
+            memcpy(new_edges + 1, body->states[s], num_orig_edges * sizeof(MVMNFAStateInfo));
+            qsort(new_edges, num_new_edges, sizeof(MVMNFAStateInfo), opt_edge_comp);
+            MVM_free(body->states[s]);
+            body->states[s] = new_edges;
+            body->num_state_edges[s] = num_new_edges;
+        }
+    }
+}
+
 /* Deserializes the data. */
 static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMSerializationReader *reader) {
     MVMNFABody *body = (MVMNFABody *)data;
@@ -161,8 +231,9 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
         body->states = MVM_malloc(body->num_states * sizeof(MVMNFAStateInfo *));
         for (i = 0; i < body->num_states; i++) {
             MVMint64 edges = body->num_state_edges[i];
-            if (edges > 0)
+            if (edges > 0) {
                 body->states[i] = MVM_malloc(edges * sizeof(MVMNFAStateInfo));
+            }
             for (j = 0; j < edges; j++) {
                 body->states[i][j].act = MVM_serialization_read_int(tc, reader);
                 body->states[i][j].to = MVM_serialization_read_int(tc, reader);
@@ -215,6 +286,8 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
             }
         }
     }
+
+    sort_states_and_add_synth_cp_node(tc, body);
 }
 
 /* Compose the representation. */
@@ -243,10 +316,10 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
 
 /* Initializes the representation. */
 const MVMREPROps * MVMNFA_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &NFA_this_repr;
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps NFA_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     NULL, /* initialize */
@@ -280,7 +353,7 @@ static const MVMREPROps this_repr = {
  * 1-char string for synthetics. */
 static MVMGrapheme32 get_grapheme(MVMThreadContext *tc, MVMObject *obj) {
     /* Handle null and non-concrete case. */
-    if (MVM_is_null(tc, obj) || !IS_CONCRETE(obj)) {
+    if (MVM_UNLIKELY(MVM_is_null(tc, obj) || !IS_CONCRETE(obj))) {
         MVM_exception_throw_adhoc(tc,
             "NFA must be provided with a concrete string or integer for graphemes");
     }
@@ -305,8 +378,7 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
     MVMNFABody *nfa;
     MVMint64    i, j, num_states;
 
-    MVMROOT(tc, states, {
-    MVMROOT(tc, nfa_type, {
+    MVMROOT2(tc, states, nfa_type, {
         /* Create NFA object. */
         nfa_obj = MVM_repr_alloc_init(tc, nfa_type);
         nfa = (MVMNFABody *)OBJECT_BODY(nfa_obj);
@@ -318,8 +390,8 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
         num_states = MVM_repr_elems(tc, states) - 1;
         nfa->num_states = num_states;
         if (num_states > 0) {
-            nfa->num_state_edges = MVM_malloc(num_states * sizeof(MVMint64));
-            nfa->states = MVM_malloc(num_states * sizeof(MVMNFAStateInfo *));
+            nfa->num_state_edges = MVM_calloc(num_states, sizeof(MVMint64));
+            nfa->states = MVM_calloc(num_states, sizeof(MVMNFAStateInfo *));
         }
         for (i = 0; i < num_states; i++) {
             MVMObject *edge_info = MVM_repr_at_pos_o(tc, states, i + 1);
@@ -328,14 +400,17 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
             MVMint64   cur_edge  = 0;
 
             nfa->num_state_edges[i] = edges;
-            if (edges > 0)
+            if (edges > 0) {
                 nfa->states[i] = MVM_malloc(edges * sizeof(MVMNFAStateInfo));
+            }
 
             for (j = 0; j < elems; j += 3) {
                 MVMint64 act = MVM_coerce_simple_intify(tc,
                     MVM_repr_at_pos_o(tc, edge_info, j));
                 MVMint64 to  = MVM_coerce_simple_intify(tc,
                     MVM_repr_at_pos_o(tc, edge_info, j + 2));
+                if (to <= 0 && act != MVM_NFA_EDGE_FATE)
+                    MVM_exception_throw_adhoc(tc, "Invalid to edge %"PRId64" in NFA statelist", to);
 
                 nfa->states[i][cur_edge].act = act;
                 nfa->states[i][cur_edge].to = to;
@@ -388,37 +463,50 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
             }
         }
     });
-    });
+
+    sort_states_and_add_synth_cp_node(tc, nfa);
 
     return nfa_obj;
 }
 
 /* Does a run of the NFA. Produces a list of integers indicating the
  * chosen ordering. */
+static MVMint32 in_done(MVMuint32 *done, MVMuint32 numdone, MVMuint32 st) {
+    MVMuint32 i = 0;
+    for (i = 0; i < numdone; i++)
+        if (done[i] == st)
+            return 1;
+    return 0;
+}
+
 static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *target, MVMint64 offset, MVMint64 *total_fates_out) {
     MVMint64  eos     = MVM_string_graphs(tc, target);
-    MVMint64  gen     = 1;
     MVMint64  numcur  = 0;
     MVMint64  numnext = 0;
-    MVMint64 *done, *fates, *curst, *nextst, *longlit;
+    MVMint64  numdone = 0;
+    MVMuint32 *done, *curst, *nextst;
+    MVMint64  *fates, *longlit;
     MVMint64  i, fate_arr_len, num_states, total_fates, prev_fates, usedlonglit;
     MVMint64  orig_offset = offset;
+    /* We used a cached grapheme iterator since we often request the same
+     * grapheme multiple times, most common after that is requesting the next
+     * grapheme. */
+    MVMGraphemeIter_cached gic;
     int nfadeb = tc->instance->nfa_debug_enabled;
 
     /* Obtain or (re)allocate "done states", "current states" and "next
      * states" arrays. */
     num_states = nfa->num_states;
     if (tc->nfa_alloc_states < num_states) {
-        size_t alloc   = (num_states + 1) * sizeof(MVMint64);
-        tc->nfa_done   = (MVMint64 *)MVM_realloc(tc->nfa_done, alloc);
-        tc->nfa_curst  = (MVMint64 *)MVM_realloc(tc->nfa_curst, alloc);
-        tc->nfa_nextst = (MVMint64 *)MVM_realloc(tc->nfa_nextst, alloc);
+        size_t alloc   = (num_states + 1) * sizeof(MVMuint32);
+        tc->nfa_done   = (MVMuint32 *)MVM_realloc(tc->nfa_done, alloc);
+        tc->nfa_curst  = (MVMuint32 *)MVM_realloc(tc->nfa_curst, alloc);
+        tc->nfa_nextst = (MVMuint32 *)MVM_realloc(tc->nfa_nextst, alloc);
         tc->nfa_alloc_states = num_states;
     }
     done   = tc->nfa_done;
     curst  = tc->nfa_curst;
     nextst = tc->nfa_nextst;
-    memset(done, 0, (num_states + 1) * sizeof(MVMint64));
 
     /* Allocate fates array. */
     fate_arr_len = 1 + MVM_repr_elems(tc, nfa->fates);
@@ -428,7 +516,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
     }
     fates = tc->nfa_fates;
     total_fates = 0;
-    if (nfadeb) fprintf(stderr,"======================================\nStarting with %d fates in %d states\n", (int)fate_arr_len, (int)num_states) ;
+    if (MVM_UNLIKELY(nfadeb)) fprintf(stderr,"======================================\nStarting with %d fates in %d states\n", (int)fate_arr_len, (int)num_states) ;
 
     /* longlit will be updated on a fate whenever NFA passes through final char of a literal. */
     /* These edges are specially marked to indicate which fate they influence the fate of. */
@@ -440,24 +528,29 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
     usedlonglit = 0;
 
     nextst[numnext++] = 1;
+    /* In NFA could be called with a string that has 0 graphemes in it. Guard
+     * against this so we don't init for the empty string. */
+    if (target->body.num_graphs) MVM_string_gi_cached_init(tc, &gic, target, 0);
+
     while (numnext && offset <= eos) {
         /* Swap next and current */
-        MVMint64 *temp = curst;
+        MVMuint32 *temp = curst;
         curst   = nextst;
         nextst  = temp;
         numcur  = numnext;
         numnext = 0;
+        numdone = 0;
 
         /* Save how many fates we have before this position is considered. */
         prev_fates = total_fates;
 
-        if (nfadeb) {
+        if (MVM_UNLIKELY(nfadeb)) {
             if (offset < eos) {
                 MVMGrapheme32 cp = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
-                fprintf(stderr,"%c with %ds target %lx offset %lld\n",cp,(int)numcur, (long)target, offset);
+                fprintf(stderr,"%c with %"PRId64"s target %"PRIXPTR" offset %"PRId64"\n", cp, numcur, (uintptr_t)target, offset);
             }
             else {
-                fprintf(stderr,"EOS with %ds\n",(int)numcur);
+                fprintf(stderr,"EOS with %"PRId64"s\n", numcur);
             }
         }
         while (numcur) {
@@ -466,15 +559,15 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
 
             MVMint64 st = curst[--numcur];
             if (st <= num_states) {
-                if (done[st] == gen)
+                if (in_done(done, numdone, st))
                     continue;
-                done[st] = gen;
+                done[numdone++] = st;
             }
 
             edge_info = nfa->states[st - 1];
             edge_info_elems = nfa->num_state_edges[st - 1];
-            if (nfadeb)
-                fprintf(stderr,"\t%d\t%d\t",(int)st, (int)edge_info_elems);
+            if (MVM_UNLIKELY(nfadeb))
+                fprintf(stderr,"\t%"PRIi64"\t%"PRIi64"\t", st, edge_info_elems);
             for (i = 0; i < edge_info_elems; i++) {
                 MVMint64 act = edge_info[i].act;
                 MVMint64 to  = edge_info[i].to;
@@ -492,7 +585,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         MVMint64 arg = edge_info[i].arg.i;
                         MVMint64 j;
                         MVMint64 found_fate = 0;
-                        if (nfadeb)
+                        if (MVM_UNLIKELY(nfadeb))
                             fprintf(stderr, "fate(%016llx) ", (long long unsigned int)arg);
                         for (j = 0; j < total_fates; j++) {
                             if (found_fate)
@@ -506,7 +599,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         total_fates -= found_fate;
                         if (arg < usedlonglit)
                             arg -= longlit[arg] << 24;
-                        if (++total_fates > fate_arr_len) {
+                        if (MVM_UNLIKELY(++total_fates > fate_arr_len)) {
                             /* should never happen if nfa->fates is correct and dedup above works right */
                             fprintf(stderr, "oops adding %016llx to\n", (long long unsigned int)arg);
                             for (j = 0; j < total_fates - 1; j++) {
@@ -526,79 +619,80 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         fates[++j] = arg;
                         continue;
                     }
-                    else if (act == MVM_NFA_EDGE_EPSILON && to <= num_states && done[to] != gen) {
-			if (to)
-			    curst[numcur++] = to;
-			else if (nfadeb)  /* XXX should turn into a "can't happen" after rebootstrap */
-			    fprintf(stderr, "  oops, ignoring epsilon to 0\n");
+                    else if (act == MVM_NFA_EDGE_EPSILON && to <= num_states &&
+                            !in_done(done, numdone, to)) {
+                        if (to)
+                            curst[numcur++] = to;
+                        else if (MVM_UNLIKELY(nfadeb))  /* XXX should turn into a "can't happen" after rebootstrap */
+                            fprintf(stderr, "  oops, ignoring epsilon to 0\n");
                         continue;
                     }
                 }
 
-                if (offset >= eos) {
+                if (eos <= offset) {
                     /* Can't match, so drop state. */
                     continue;
                 }
                 else {
                     switch (act) {
                         case MVM_NFA_EDGE_CODEPOINT_LL: {
-                            MVMGrapheme32 arg = edge_info[i].arg.g;
-                            if (MVM_string_get_grapheme_at_nocheck(tc, target, offset) == arg) {
+                            const MVMGrapheme32 arg = edge_info[i].arg.g;
+                            if (MVM_string_gi_cached_get_grapheme(tc, &gic, offset) == arg) {
                                 MVMint64 fate = (edge_info[i].act >> 8) & 0xfffff;
                                 nextst[numnext++] = to;
                                 while (usedlonglit <= fate)
                                     longlit[usedlonglit++] = 0;
                                 longlit[fate] = offset - orig_offset + 1;
-                                if (nfadeb)
+                                if (MVM_UNLIKELY(nfadeb))
                                     fprintf(stderr, "%d->%d ", (int)i, (int)to);
                             }
                             continue;
                         }
                         case MVM_NFA_EDGE_CODEPOINT: {
-                            MVMGrapheme32 arg = edge_info[i].arg.g;
-                            if (MVM_string_get_grapheme_at_nocheck(tc, target, offset) == arg) {
+                            const MVMGrapheme32 arg = edge_info[i].arg.g;
+                            if (MVM_string_gi_cached_get_grapheme(tc, &gic, offset) == arg) {
                                 nextst[numnext++] = to;
-                                if (nfadeb)
+                                if (MVM_UNLIKELY(nfadeb))
                                     fprintf(stderr, "%d->%d ", (int)i, (int)to);
                             }
                             continue;
                         }
                         case MVM_NFA_EDGE_CODEPOINT_NEG: {
-                            MVMGrapheme32 arg = edge_info[i].arg.g;
-                            if (MVM_string_get_grapheme_at_nocheck(tc, target, offset) != arg)
+                            const MVMGrapheme32 arg = edge_info[i].arg.g;
+                            if (MVM_string_gi_cached_get_grapheme(tc, &gic, offset) != arg)
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARCLASS: {
-                            MVMint64 arg = edge_info[i].arg.i;
-                            if (MVM_string_is_cclass(tc, arg, target, offset))
+                            const MVMint64 arg = edge_info[i].arg.i;
+                            if (MVM_string_grapheme_is_cclass(tc, arg, MVM_string_gi_cached_get_grapheme(tc, &gic, offset)))
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARCLASS_NEG: {
-                            MVMint64 arg = edge_info[i].arg.i;
-                            if (!MVM_string_is_cclass(tc, arg, target, offset))
+                            const MVMint64 arg = edge_info[i].arg.i;
+                            if (!MVM_string_grapheme_is_cclass(tc, arg, MVM_string_gi_cached_get_grapheme(tc, &gic, offset)))
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARLIST: {
-                            MVMString *arg    = edge_info[i].arg.s;
-                            MVMGrapheme32 cp = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
+                            MVMString *arg   = edge_info[i].arg.s;
+                            MVMGrapheme32 cp = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
                             if (MVM_string_index_of_grapheme(tc, arg, cp) >= 0)
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARLIST_NEG: {
                             MVMString *arg    = edge_info[i].arg.s;
-                            MVMGrapheme32 cp = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
+                            const MVMGrapheme32 cp = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
                             if (MVM_string_index_of_grapheme(tc, arg, cp) < 0)
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CODEPOINT_I_LL: {
-                            MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
-                            MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
+                            const MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
+                            const MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
+                            const MVMGrapheme32 ord    = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
                             if (ord == lc_arg || ord == uc_arg) {
                                 MVMint64 fate = (edge_info[i].act >> 8) & 0xfffff;
                                 nextst[numnext++] = to;
@@ -611,15 +705,15 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         case MVM_NFA_EDGE_CODEPOINT_I: {
                             MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
                             MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
+                            MVMGrapheme32 ord    = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
                             if (ord == lc_arg || ord == uc_arg)
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CODEPOINT_I_NEG: {
-                            MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
-                            MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
+                            const MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
+                            const MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
+                            const MVMGrapheme32 ord    = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
                             if (ord != lc_arg && ord != uc_arg)
                                 nextst[numnext++] = to;
                             continue;
@@ -627,21 +721,21 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         case MVM_NFA_EDGE_CHARRANGE: {
                             MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
                             MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
+                            MVMGrapheme32 ord    = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
                             if (ord >= lc_arg && ord <= uc_arg)
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARRANGE_NEG: {
-                            MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
-                            MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
-                            if (ord < lc_arg || ord > uc_arg)
+                            const MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
+                            const MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
+                            const MVMGrapheme32 ord    = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
+                            if (ord < lc_arg || uc_arg < ord)
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_SUBRULE:
-                            if (nfadeb)
+                            if (MVM_UNLIKELY(nfadeb))
                                 fprintf(stderr, "IGNORING SUBRULE\n");
                             continue;
                         case MVM_NFA_EDGE_CODEPOINT_M:
@@ -669,7 +763,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             MVMint32 ready;
                             MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
                             MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_ord_basechar_at(tc, target, offset);
+                            const MVMGrapheme32 ord = MVM_string_ord_basechar_at(tc, target, offset);
 
                             MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
                             ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, uc_arg, &uc_arg);
@@ -691,36 +785,88 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARRANGE_M: {
-                            MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
-                            MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_ord_basechar_at(tc, target, offset);
+                            const MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
+                            const MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
+                            const MVMGrapheme32 ord    = MVM_string_ord_basechar_at(tc, target, offset);
                             if (ord >= lc_arg && ord <= uc_arg)
                                 nextst[numnext++] = to;
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARRANGE_M_NEG: {
-                            MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
-                            MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
-                            MVMGrapheme32 ord    = MVM_string_ord_basechar_at(tc, target, offset);
-                            if (ord < lc_arg || ord > uc_arg)
+                            const MVMGrapheme32 uc_arg = edge_info[i].arg.uclc.uc;
+                            const MVMGrapheme32 lc_arg = edge_info[i].arg.uclc.lc;
+                            const MVMGrapheme32 ord    = MVM_string_ord_basechar_at(tc, target, offset);
+                            if (ord < lc_arg || uc_arg < ord)
                                 nextst[numnext++] = to;
                             continue;
                         }
+                        case MVM_NFA_EDGE_SYNTH_CP_COUNT: {
+                            /* Binary search the edges ahead for the grapheme. */
+                            const MVMGrapheme32 search = MVM_string_gi_cached_get_grapheme(tc, &gic, offset);
+                            const MVMint64 num_possibilities = edge_info[i].arg.i;
+                            const MVMint64 end = i + num_possibilities;
+                            MVMint64 l = i + 1;
+                            MVMint64 r = end;
+                            MVMint64 found = -1;
+                            while (l <= r) {
+                                const MVMint64 m = l + (r - l) / 2;
+                                const MVMGrapheme32 test = edge_info[m].arg.g;
+                                if (test == search) {
+                                    /* We found it, but important we get the first edge
+                                     * that matches. */
+                                    found = m;
+                                    while (found > i + 1 && edge_info[found - 1].arg.g == search)
+                                        found--;
+                                    break;
+                                }
+                                if (test < search)
+                                    l = m + 1;
+                                else
+                                    r = m - 1;
+                            }
+                            if (found == -1) {
+                                /* Binary search failed to find a match, so just skip all
+                                 * the nodes. */
+                                i += num_possibilities;
+                            }
+                            else {
+                                /* Add all states that match. */
+                                while (found <= end && edge_info[found].arg.g == search) {
+                                    to = edge_info[found].to;
+                                    if (edge_info[found].act == MVM_NFA_EDGE_CODEPOINT) {
+                                        nextst[numnext++] = to;
+                                        if (MVM_UNLIKELY(nfadeb))
+                                            fprintf(stderr, "%d->%d ", (int)found, (int)to);
+                                    }
+                                    else {
+                                        const MVMint64 fate = (edge_info[found].act >> 8) & 0xfffff;
+                                        nextst[numnext++] = to;
+                                        while (usedlonglit <= fate)
+                                            longlit[usedlonglit++] = 0;
+                                        longlit[fate] = offset - orig_offset + 1;
+                                        if (MVM_UNLIKELY(nfadeb))
+                                            fprintf(stderr, "%d->%d ", (int)found, (int)to);
+                                    }
+                                    found++;
+                                }
+                                i += num_possibilities;
+                            }
+                            break;
+                        }
                     }
                 }
             }
-            if (nfadeb) fprintf(stderr,"\n");
+            if (MVM_UNLIKELY(nfadeb)) fprintf(stderr,"\n");
         }
 
-        /* Move to next character and generation. */
+        /* Move to next character. */
         offset++;
-        gen++;
     }
     /* strip any literal lengths, leaving only fates */
     if (usedlonglit || nfadeb) {
-        if (nfadeb) fprintf(stderr,"Final\n");
+        if (MVM_UNLIKELY(nfadeb)) fprintf(stderr,"Final\n");
         for (i = 0; i < total_fates; i++) {
-            if (nfadeb) fprintf(stderr, "  %08llx\n", (long long unsigned int)fates[i]);
+            if (MVM_UNLIKELY(nfadeb)) fprintf(stderr, "  %08llx\n", (long long unsigned int)fates[i]);
             fates[i] &= 0xffffff;
         }
     }
