@@ -4,6 +4,35 @@
 #if defined(_MSC_VER)
 #define strtoll _strtoi64
 #define snprintf _snprintf
+/* slightly adapted to use in MoarVM */
+//===-- int_lib.h - configuration header for compiler-rt  -----------------===//
+//
+// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
+// See https://llvm.org/LICENSE.txt for license information.
+// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
+//
+//===----------------------------------------------------------------------===//
+//
+// This file is a configuration header for compiler-rt.
+// This file is not part of the interface of this library.
+//
+//===----------------------------------------------------------------------===//
+#if defined(_M_ARM) || defined(_M_X64)
+static int __inline __builtin_clzll(uint64_t value) {
+  unsigned long leading_zero = 0;
+  if (_BitScanReverse64(&leading_zero, value))
+    return 63 - leading_zero;
+  return 63;
+}
+#else
+static int __inline __builtin_clzll(uint64_t value) {
+  uint32_t msh = (uint32_t)(value >> 32);
+  uint32_t lsh = (uint32_t)(value & 0xFFFFFFFF);
+  if (msh != 0)
+    return __builtin_clz(msh);
+  return 32 + __builtin_clz(lsh);
+}
+#endif
 #endif
 
 MVMint64 MVM_coerce_istrue_s(MVMThreadContext *tc, MVMString *str) {
@@ -34,20 +63,25 @@ LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE. */
 
-#define W(N, I) b[N] = (char)(I) + '0'
-#define A(N) t = ((uint64_t)(1) << (32 + N / 5 * N * 53 / 16)) / (uint32_t)(1e##N) + 1 - N / 9, t *= u, t >>= N / 5 * N * 53 / 16, t += N / 5 * 4, W(0, t >> 32)
-#define D(N) t = (uint64_t)(10) * (uint32_t)(t), W(N, t >> 32)
+typedef struct pair { char t, o; } pair;
+#define P(T) T, '0',  T, '1', T, '2', T, '3', T, '4', T, '5', T, '6', T, '7', T, '8', T, '9'
+static const pair s_pairs[] = { P('0'), P('1'), P('2'), P('3'), P('4'), P('5'), P('6'), P('7'), P('8'), P('9') };
 
-#define L0 W(0, u)
-#define L1 A(1), D(1)
-#define L2 A(2), D(1), D(2)
-#define L3 A(3), D(1), D(2), D(3)
-#define L4 A(4), D(1), D(2), D(3), D(4)
-#define L5 A(5), D(1), D(2), D(3), D(4), D(5)
-#define L6 A(6), D(1), D(2), D(3), D(4), D(5), D(6)
-#define L7 A(7), D(1), D(2), D(3), D(4), D(5), D(6), D(7)
-#define L8 A(8), D(1), D(2), D(3), D(4), D(5), D(6), D(7), D(8)
-#define L9 A(9), D(1), D(2), D(3), D(4), D(5), D(6), D(7), D(8), D(9)
+#define W(N, I) *(pair*)&b[N] = s_pairs[I]
+#define A(N) t = ((uint64_t)(1) << (32 + N / 5 * N * 53 / 16)) / (uint32_t)(1e##N) + 1 + N/6 - N/8, t *= u, t >>= N / 5 * N * 53 / 16, t += N / 6 * 4, W(0, t >> 32)
+#define S(N) b[N] = (char)((uint64_t)(10) * (uint32_t)(t) >> 32) + '0'
+#define D(N) t = (uint64_t)(100) * (uint32_t)(t), W(N, t >> 32)
+
+#define L0 b[0] = (char)(u) + '0'
+#define L1 W(0, u)
+#define L2 A(1), S(2)
+#define L3 A(2), D(2)
+#define L4 A(3), D(2), S(4)
+#define L5 A(4), D(2), D(4)
+#define L6 A(5), D(2), D(4), S(6)
+#define L7 A(6), D(2), D(4), D(6)
+#define L8 A(7), D(2), D(4), D(6), S(8)
+#define L9 A(8), D(2), D(4), D(6), D(8)
 
 #define LN(N) (L##N, b += N + 1)
 #define LZ LN
@@ -84,55 +118,52 @@ static char * i64toa_jeaiii(int64_t i, char* b) {
 }
 
 /* End code */
-
+static const int mag[] = { 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 15, 15, 15, 16, 16, 16, 16, 17, 17, 17, 18, 18, 18, 19, 19, 19, 19, 20 };
 MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
-    char buffer[20];
-    int len;
     /* See if we can hit the cache. */
-    int cache = 0 <= i && i < MVM_INT_TO_STR_CACHE_SIZE;
+    const int cache = 0 <= i && i < MVM_INT_TO_STR_CACHE_SIZE;
     if (cache) {
         MVMString *cached = tc->instance->int_to_str_cache[i];
         if (cached)
             return cached;
     }
     /* Otherwise, need to do the work; cache it if in range. */
-    len = i64toa_jeaiii(i, buffer) - buffer;
+    const int is_negative = i < 0;
+    const int msb = 64 - __builtin_clzll((is_negative ? -i : i) | 1);
+    char *buffer = MVM_malloc(mag[msb] + is_negative);
+    const int len = i64toa_jeaiii(i, buffer) - buffer;
     if (0 <= len) {
-        MVMString *result = NULL;
-        MVMGrapheme8 *blob = MVM_malloc(len);
-        memcpy(blob, buffer, len);
-        result = MVM_string_ascii_from_buf_nocheck(tc, blob, len);
+        MVMString *result = MVM_string_ascii_from_buf_nocheck(tc, (MVMGrapheme8 *)buffer, len);
         if (cache)
             tc->instance->int_to_str_cache[i] = result;
         return result;
     }
     else {
+        MVM_free(buffer);
         MVM_exception_throw_adhoc(tc, "Could not stringify integer (%"PRId64")", i);
     }
 }
 
 MVMString * MVM_coerce_u_s(MVMThreadContext *tc, MVMuint64 i) {
-    char buffer[20];
-    int len;
     /* See if we can hit the cache. */
-    int cache = i < MVM_INT_TO_STR_CACHE_SIZE;
+    const int cache = i < MVM_INT_TO_STR_CACHE_SIZE;
     if (cache) {
         MVMString *cached = tc->instance->int_to_str_cache[i];
         if (cached)
             return cached;
     }
     /* Otherwise, need to do the work; cache it if in range. */
-    len = u64toa_jeaiii(i, buffer) - buffer;
+    const int msb = 64 - __builtin_clzll(i | 1);
+    char *buffer = MVM_malloc(mag[msb]);
+    const int len = u64toa_jeaiii(i, buffer) - buffer;
     if (0 <= len) {
-        MVMString *result = NULL;
-        MVMGrapheme8 *blob = MVM_malloc(len);
-        memcpy(blob, buffer, len);
-        result = MVM_string_ascii_from_buf_nocheck(tc, blob, len);
+        MVMString *result = MVM_string_ascii_from_buf_nocheck(tc, (MVMGrapheme8 *)buffer, len);
         if (cache)
             tc->instance->int_to_str_cache[i] = result;
         return result;
     }
     else {
+        MVM_free(buffer);
         MVM_exception_throw_adhoc(tc, "Could not stringify integer (%"PRIu64")", i);
     }
 }
