@@ -79,10 +79,14 @@ static void async_handler(uv_async_t *handle) {
 }
 
 /* Enters the event loop. */
-static void enter_loop(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *args) {
+static void enter_loop(MVMThreadContext *tc, MVMArgs arg_info) {
     uv_loop_t    *loop = tc->instance->event_loop;
     uv_async_t   *async = tc->instance->event_loop_wakeup;
 
+#ifdef MVM_HAS_PTHREAD_SETNAME_NP
+    pthread_setname_np(pthread_self(), "async io thread");
+#endif
+
     /* Bind the thread context for the wakeup signal */
     async->data = tc;
 
@@ -296,11 +300,8 @@ void MVM_io_eventloop_destroy(MVMThreadContext *tc) {
         /* Not sure we can always do this */
         uv_loop_close(instance->event_loop);
        
-        MVM_free(instance->event_loop_wakeup);
-        instance->event_loop_wakeup = NULL;
-        
-        MVM_free(instance->event_loop);
-        instance->event_loop = NULL;
+        MVM_free_null(instance->event_loop_wakeup);
+        MVM_free_null(instance->event_loop);
     }
 
     uv_mutex_unlock(&instance->mutex_event_loop);
